from maxar_canvas_grid.grid import Cell
from collections import defaultdict
from typing import List
import warnings

from shapely.geometry import mapping, MultiPolygon


class GeoMixin():
    ''' Adds Python Geospatial Interface to objects with tiles '''

    @property
    def __geo_interface__(self) -> dict:
        geoms = [tile.geom_WGS84 for tile in self.tiles]
        return mapping(MultiPolygon(geoms)) 




class Acquisition(list, GeoMixin):
    ''' Container for one or more tiles from a single acquisition 
    
        These are generated by ARDCollections and would not normally be initialized '''

    def __init__(self, acq_id: str, tiles: list = []) -> None:

        self.acq_id = acq_id
        ''' str: ID of this acquisition '''
        super().__init__(*tiles)

    @property
    def cells(self) -> List[Cell]:
        cells = set([tile.cell for tile in self])
        return list(cells)

    def get_tile_from_cell(self, cell):
        cell = Cell(cell)
        for tile in self:
            if tile.cell == cell:
                return tile
        raise ValueError('Tile not found in this acquisition')

    @property
    def tiles(self):
        ''' not needed but provides access to tiles for GeoMixin '''
        return self

    def __str__(self):
        return f'<Acquisition at {self.acq_id} ({len(self)} tiles)>'

    def __repr__(self):
        tiles = ', '.join([f'<{type(tile).__name__} at {tile.cell.id}>' for tile in self])
        return f'<Acquisition of {self.acq_id} [{tiles}]>'

    def __hash__(self):
        ''' Hash of cell ID '''
        return hash(self.acq_id) 

    @property
    def properties(self) -> dict:
        ''' Strip level properties '''

        props = self[0].properties
        # todo: check if catalog_id is still passed in metadata
        keep = ['catalog_id', 'acquisition_id', 'platform', 'scan_direction','epsg','datetime']
        strip_props = {}
        for key in keep:
            try:
                strip_props[key] = props[key]
            except KeyError:
                pass 
        return strip_props


            
class Stack(list, GeoMixin):
    ''' Container for one or more tiles in an ARD grid cell
    
        These are generated by ARDCollections and would not normally be initialized '''

    def __init__(self, cell: Cell, tiles: list = []) -> None:

        self.cell = Cell(cell)
        super().__init__(*tiles) 

    @property
    def tiles(self):
        ''' not needed but provides access to tiles for GeoMixin '''
        return self

    def get_tile_from_acquisition(self, acq_id):
        for tile in self:
            if tile.acq_id == acq_id:
                return tile
        raise ValueError('Tile not found in this stack')


    @property
    def acquisition_ids(self) -> List[str]:
        ids = set([tile.acq_id for tile in self])
        return list(ids)

    def __str__(self):
        return f'<Stack at {self.cell.id} ({len(self)} tiles)>'

    def __repr__(self):
        tiles = ', '.join([f'<{type(tile).__name__} of {tile.acq_id}>' for tile in self])
        return f'<Stack at {self.cell.id} [{tiles}]>'

    def __hash__(self):
        ''' Hash of cell ID '''
        return hash(self.cell) 


class Store(dict):
    ''' A defaultdict-like k:v store for objects '''

    def __init__(self, object_type, *args, **kwargs):
        self.object_type = object_type
        super().__init__(*args, **kwargs)

    def __missing__(self, key):
        res = self[key] = self.object_type(key)
        return res

class BaseCollection(GeoMixin):
    
    def __init__(self):
        self._reset()

    def _reset(self):
        self._acquisitions = Store(Acquisition)
        self._stacks = Store(Stack)
        self._zones = set()
        self._dates = set()

    def add_tile(self, tile):
        self._acquisitions[tile.acq_id].append(tile)
        self._stacks[tile.cell].append(tile)
        self._zones.add(tile.cell.zone)
        self._dates.add(tile.date)

    @property
    def tiles(self):
        return [tile for tiles in self.stacks for tile in tiles]

    def get_tile(self, acq_id, cell):
        cell = Cell(cell)
        acq  = self.get_acquisition(acq_id)
        for tile in acq:
            if tile.cell == cell:
                return tile
        raise ValueError('Tile not found in this collection')

    @property
    def dates(self) -> List[str]:
        ''' All of the dates of acquisitions in the ARDCollection '''
        dates = list(self._dates)
        dates.sort()
        return dates

    @property
    def start_date(self) -> str:
        ''' Earliest date of acquisitions in the ARDCollection as YYYY-MM-DD '''
        return self.dates[0]

    @property
    def end_date(self) -> str:
        # does the type get caught?
        ''' Latest date of acquisitions in the ARDCollection as YYYY-MM-DD '''
        return self.dates[-1]

    @property
    def zones(self) -> List[int]:
        ''' The UTM zones covered by tiles in the ARDCollection '''
        return list(self._zones)

    #
    # Stacks
    #

    @property
    def stacks(self):
        return list(self._stacks.values())

    @property
    def cells(self):
        return list(self._stacks.keys())

    def get_stack(self, cell):
        cell = Cell(cell)
        try:
            return self._stacks[cell]
        except KeyError:
            raise ValueError(f'Stack for cell {cell} not found in this collection')
    #
    # Acquisitions
    #
    
    @property
    def acquisitions(self):
        return list(self._acquisitions.values())
    
    @property
    def acquisition_ids(self):
        return list(self._acquisitions.keys())
    
    @property
    def acq_ids(self):
        return self.acquisition_ids
 
    def get_acquisition(self, acq_id):
        try:
            return self._acquisitions[acq_id]
        except KeyError:
            raise ValueError(f'Aquisition for ID {acq_id} not found in this collection')


class BaseTile():
    ''' Not used yet, but a base class / ABC might be good for tiles '''

    def __init__(self, acq_id, cell):
        self.cell = cell
        self.acq_id = acq_id


    @property
    def __geo_interface__(self) -> dict:
        geoms = [tile.geom_WGS84 for tile in self.tiles]
        return mapping(MultiPolygon(geoms)) 